class FutureAccessError(Exception):
    pass


class SerializationError(Exception):
    pass


class BatchTaskError(Exception):
    pass


class TaskOutsError(Exception):
    pass


class EvalError(Exception):
    pass
